#!/usr/bin/env python3
## irccat.py
# Simple threaded irccat implementation, using pydle.
import sys
import os
import threading
import logging
import asyncio
from asyncio.streams import FlowControlMixin

from .. import  Client, __version__
from . import _args
import asyncio

class IRCCat(Client):
    """ irccat. Takes raw messages on stdin, dumps raw messages to stdout. Life has never been easier. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.async_stdin = None

    @asyncio.coroutine
    def _send(self, data):
        sys.stdout.write(data)
        yield from super()._send(data)

    @asyncio.coroutine
    def process_stdin(self):
        """ Yes. """
        loop = self.eventloop.loop

        self.async_stdin = asyncio.StreamReader()
        reader_protocol = asyncio.StreamReaderProtocol(self.async_stdin)
        yield from loop.connect_read_pipe(lambda: reader_protocol, sys.stdin)

        while True:
            line = yield from self.async_stdin.readline()
            if not line:
                break
            yield from self.raw(line.decode('utf-8'))

        yield from self.quit('EOF')

    @asyncio.coroutine
    def on_raw(self, message):
        print(message._raw)
        yield from super().on_raw(message)

    @asyncio.coroutine
    def on_ctcp_version(self, source, target, contents):
        self.ctcp_reply(source, 'VERSION', 'pydle-irccat v{}'.format(__version__))


def main():
    # Setup logging.
    logging.basicConfig(format='!! %(levelname)s: %(message)s')

    # Create client.
    irccat, connect = _args.client_from_args('irccat', default_nick='irccat', description='Process raw IRC messages from stdin, dump received IRC messages to stdout.', cls=IRCCat)

    irccat.eventloop.schedule_async(connect())
    irccat.eventloop.run_with(irccat.process_stdin())


if __name__ == '__main__':
    main()



## whox.py
# WHOX support.
from pydle.features import isupport, account

NO_ACCOUNT = '0'
# Maximum of 3 characters because Charybdis stupidity. The ASCII values of 'pydle' added together.
WHOX_IDENTIFIER = '542'

class WHOXSupport(isupport.ISUPPORTSupport, account.AccountSupport):

    ## Overrides.

    async def on_raw_join(self, message):
        """ Override JOIN to send WHOX. """
        await super().on_raw_join(message)
        nick, metadata = self._parse_user(message.source)
        channels = message.params[0].split(',')

        if self.is_same_nick(self.nickname, nick):
            # We joined.
            if 'WHOX' in self._isupport and self._isupport['WHOX']:
                # Get more relevant channel info thanks to WHOX.
                await self.rawmsg('WHO', ','.join(channels), '%tnurha,{id}'.format(id=WHOX_IDENTIFIER))
        else:
            # Find account name of person.
            pass
    
    def _create_user(self, nickname):
        super()._create_user(nickname)
        if self.registered and 'WHOX' not in self._isupport:
            self.whois(nickname)

    async def on_raw_354(self, message):
        """ WHOX results have arrived. """
        # Is the message for us?
        target, identifier = message.params[:2]
        if identifier != WHOX_IDENTIFIER:
            return

        # Great. Extract relevant information.
        metadata = {
            'nickname': message.params[4],
            'username': message.params[2],
            'realname': message.params[6],
            'hostname': message.params[3],
        }
        if message.params[5] != NO_ACCOUNT:
            metadata['identified'] = True
            metadata['account'] = message.params[5]

        self._sync_user(metadata['nickname'], metadata)



from . import cap

VISIBLITY_ALL = '*'


class MetadataSupport(cap.CapabilityNegotiationSupport):

    ## Internals.

    def _reset_attributes(self):
        super()._reset_attributes()

        self._pending['metadata'] = {}
        self._metadata_info = {}
        self._metadata_queue = []

    ## IRC API.

    async def get_metadata(self, target):
        """
        Return user metadata information.
        This is a blocking asynchronous method: it has to be called from a coroutine, as follows:

            metadata = await self.get_metadata('#foo')
        """
        if target not in self._pending['metadata']:
            await self.rawmsg('METADATA', target, 'LIST')

            self._metadata_queue.append(target)
            self._metadata_info[target] = {}
            self._pending['metadata'][target] = self.eventloop.create_future()

        return self._pending['metadata'][target]

    async def set_metadata(self, target, key, value):
        await self.rawmsg('METADATA', target, 'SET', key, value)

    async def unset_metadata(self, target, key):
        await self.rawmsg('METADATA', target, 'SET', key)

    async def clear_metadata(self, target):
        await self.rawmsg('METADATA', target, 'CLEAR')


    ## Callbacks.

    async def on_metadata(self, target, key, value, visibility=None):
        pass


    ## Message handlers.

    async def on_capability_metadata_notify_available(self, value):
        return True

    async def on_raw_metadata(self, message):
        """ Metadata event. """
        target, targetmeta = self._parse_user(message.params[0])
        key, visibility, value = message.params[1:4]
        if visibility == VISIBLITY_ALL:
            visibility = None

        if target in self.users:
            self._sync_user(target, targetmeta)
        await self.on_metadata(target, key, value, visibility=visibility)

    async def on_raw_760(self, message):
        """ Metadata key/value for whois. """
        target, targetmeta = self._parse_user(message.params[0])
        key, _, value = message.params[1:4]

        if target not in self._pending['whois']:
            return
        if target in self.users:
            self._sync_user(target, targetmeta)

        self._whois_info[target].setdefault('metadata', {})
        self._whois_info[target]['metadata'][key] = value

    async def on_raw_761(self, message):
        """ Metadata key/value. """
        target, targetmeta = self._parse_user(message.params[0])
        key, visibility = message.params[1:3]
        value = message.params[3] if len(message.params) > 3 else None

        if target not in self._pending['metadata']:
            return
        if target in self.users:
            self._sync_user(target, targetmeta)

        self._metadata_info[target][key] = value

    async def on_raw_762(self, message):
        """ End of metadata. """
        # No way to figure out whose query this belongs to, so make a best guess
        # it was the first one.
        if not self._metadata_queue:
            return
        nickname = self._metadata_queue.pop()

        future = self._pending['metadata'].pop(nickname)
        future.set_result(self._metadata_info.pop(nickname))

    async def on_raw_764(self, message):
        """ Metadata limit reached. """
        pass

    async def on_raw_765(self, message):
        """ Invalid metadata target. """
        target, targetmeta = self._parse_user(message.params[0])

        if target not in self._pending['metadata']:
            return
        if target in self.users:
            self._sync_user(target, targetmeta)

        self._metadata_queue.remove(target)
        del self._metadata_info[target]

        future = self._pending['metadata'].pop(target)
        future.set_result(None)

    async def on_raw_766(self, message):
        """ Unknown metadata key. """
        pass

    async def on_raw_767(self, message):
        """ Invalid metadata key. """
        pass

    async def on_raw_768(self, message):
        """ Metadata key not set. """
        pass

    async def on_raw_769(self, message):
        """ Metadata permission denied. """
        pass



from setuptools import setup

setup(
    name='pydle',
    version='0.9.1',
    python_requires=">=3.5",
    packages=[
        'pydle',
        'pydle.features',
        'pydle.features.rfc1459',
        'pydle.features.ircv3',
        'pydle.utils'
    ],
    extras_require={
        'sasl': 'pure-sasl >=0.1.6',   # for pydle.features.sasl
        'docs': 'sphinx_rtd_theme',    # the Sphinx theme we use
        'tests': 'pytest',             # collect and run tests
        'coverage': 'pytest-cov'       # get test case coverage
    },
    entry_points={
        'console_scripts': [
            'pydle = pydle.utils.run:main',
            'pydle-irccat = pydle.utils.irccat:main'
        ]
    },

    author='Shiz',
    author_email='hi@shiz.me',
    url='https://github.com/Shizmob/pydle',
    keywords='irc library python3 compact flexible',
    description='A compact, flexible and standards-abiding IRC library for Python 3.',
    license='BSD',

    zip_safe=True,
    test_suite='tests'
)



## ircv3_1.py
# IRCv3.1 full spec support.
from pydle.features import account, tls
from . import cap
from . import sasl

__all__ = [ 'IRCv3_1Support' ]


NO_ACCOUNT = '*'

class IRCv3_1Support(sasl.SASLSupport, cap.CapabilityNegotiationSupport, account.AccountSupport, tls.TLSSupport):
    """ Support for IRCv3.1's base and optional extensions. """

    def _rename_user(self, user, new):
        # If the server supports account-notify, we will be told about the registration status changing.
        # As such, we can skip the song and dance pydle.features.account does.
        if self._capabilities.get('account-notify', False):
            account = self.users.get(user, {}).get('account', None)
            identified = self.users.get(user, {}).get('identified', False)

        super()._rename_user(user, new)

        if self._capabilities.get('account-notify', False):
            self._sync_user(new, {'account': account, 'identified': identified})

    ## IRC callbacks.

    async def on_capability_account_notify_available(self, value):
        """ Take note of user account changes. """
        return True

    async def on_capability_away_notify_available(self, value):
        """ Take note of AWAY messages. """
        return True

    async def on_capability_extended_join_available(self, value):
        """ Take note of user account and realname on JOIN. """
        return True

    async def on_capability_multi_prefix_available(self, value):
        """ Thanks to how underlying client code works we already support multiple prefixes. """
        return True

    async def on_capability_tls_available(self, value):
        """ We never need to request this explicitly. """
        return False


    ## Message handlers.

    async def on_raw_account(self, message):
        """ Changes in the associated account for a nickname. """
        if not self._capabilities.get('account-notify', False):
            return

        nick, metadata = self._parse_user(message.source)
        account = message.params[0]

        if nick not in self.users:
            return

        self._sync_user(nick, metadata)
        if account == NO_ACCOUNT:
            self._sync_user(nick, { 'account': None, 'identified': False })
        else:
            self._sync_user(nick, { 'account': account, 'identified': True })

    async def on_raw_away(self, message):
        """ Process AWAY messages. """
        if 'away-notify' not in self._capabilities or not self._capabilities['away-notify']:
            return

        nick, metadata = self._parse_user(message.source)
        if nick not in self.users:
            return

        self._sync_user(nick, metadata)
        self.users[nick]['away'] = len(message.params) > 0
        self.users[nick]['away_message'] = message.params[0] if len(message.params) > 0 else None

    async def on_raw_join(self, message):
        """ Process extended JOIN messages. """
        if 'extended-join' in self._capabilities and self._capabilities['extended-join']:
            nick, metadata = self._parse_user(message.source)
            channels, account, realname = message.params

            self._sync_user(nick, metadata)

            # Emit a fake join message.
            fakemsg = self._create_message('JOIN', channels, source=message.source)
            await super().on_raw_join(fakemsg)

            if account == NO_ACCOUNT:
                account = None
            self.users[nick]['account'] = account
            self.users[nick]['realname'] = realname
        else:
            await super().on_raw_join(message)
